
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <GL/glew.h> // http://glew.sourceforge.net/basic.html

#include <GL/glfw.h> // http://www.glfw.org/GLFWReference27.pdf

// http://www.khronos.org/files/opengl-quick-reference-card.pdf
#include "shaders.h"
#include "geometry.h"
#include "textures.h"
#include "fbo.h"

#ifndef M_PI
#define M_PI 3.141592654
#endif

GLuint winWidth=768;
GLuint winHeight=768;

float angleX=0.0;
float angleY=0.0;

int running=1;

//-------------- geometry --------------

SModel modelCube;

//-------------- textures --------------

// Texture IDs


// ------------- shaders --------------
GLuint shVertPosID=0;

float gLightPos[3];
float gAngleVal;

GLhandleARB gGLSLprog=0;

void onInit()
{
// load texture ...

// transfer func ...

// load shaders
    gGLSLprog=createProgram("shaders/minimal.vert","shaders/minimal.frag");

	GLuint shVolumeMapID=glGetUniformLocation(gGLSLprog,"volumeMap");

    glUseProgramObjectARB(gGLSLprog);

    glUniform1i(shVolumeMapID,0);

    glUseProgramObjectARB(0);

	shVertPosID=glGetAttribLocation(gGLSLprog,"inVertPosMS");

// load geometry
	generateCubeGeometry(&modelCube);

	glGenBuffers(1, &modelCube.vPosVB); // positions
	glBindBuffer(GL_ARRAY_BUFFER, modelCube.vPosVB);
    glBufferData(GL_ARRAY_BUFFER, modelCube.vertCnt*3*sizeof(float), modelCube.vPosArr,GL_STATIC_DRAW_ARB);
	delete [] modelCube.vPosArr;

	glGenBuffers(1, &modelCube.triangleIB); // Index Buffer
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, modelCube.triangleIB);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, modelCube.indexCnt*sizeof(unsigned int), modelCube.indexArr,GL_STATIC_DRAW_ARB);
	delete [] modelCube.indexArr;
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0);
}

void renderSquareTex3D(GLuint texID, float z)
{
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_3D, texID);
	glEnable(GL_TEXTURE_3D);

	glColor4f(1.0,1.0,1.0,1.0);
	glBegin(GL_QUADS);
		glTexCoord3f(0.0, 0.0, z);
		glVertex2f(-1.0, -1.0);
		glTexCoord3f(1.0, 0.0, z);
		glVertex2f(1.0, -1.0);
		glTexCoord3f(1.0, 1.0, z);
		glVertex2f(1.0, 1.0);
		glTexCoord3f(0.0, 1.0, z);
		glVertex2f(-1.0, 1.0);
	glEnd();

	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_3D);
}

void renderModel(SModel& model)
{
// setup textures ...
	
// setup shader
    glUseProgramObjectARB(gGLSLprog);

// transform lightpos & viewpos into Model Space !!!

// set view position in Model Space ...

// set ray step length ...

// set light position in Model Space

// bind VBO - positions
	glBindBuffer( GL_ARRAY_BUFFER, model.vPosVB);	// bind ARRAY BUFFER with glBuffer of vertex positions

	glVertexAttribPointerARB(shVertPosID,3,GL_FLOAT,GL_FALSE,0,0);
	glEnableVertexAttribArrayARB(shVertPosID);

// !!! OpenGL < 3.0 requires GL_VERTEX_ARRAY enabled to render geometry
	glVertexPointer(3,GL_FLOAT,0,0);
	glEnableClientState(GL_VERTEX_ARRAY);

// bind IB
	glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, model.triangleIB);	// bind ELEMENT ARRAY BUFFER with glBuffer of indices
// render	
	glDrawElements(GL_TRIANGLES,model.indexCnt,GL_UNSIGNED_INT,0);
// clean-up

	glBindBufferARB( GL_ELEMENT_ARRAY_BUFFER_ARB, 0);	// unbind ELEMENT ARRAY BUFFER
	glDisableClientState(GL_VERTEX_ARRAY);

	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_3D);
}

/* handle rendering */
void onRender()
{
	GLfloat xmax, znear, zfar, h;

// animation
	gLightPos[0]=2.0*sin(gAngleVal);
	gLightPos[1]=0.0;
	gLightPos[2]=2.0*cos(gAngleVal);

	gAngleVal+=0.01;

    glViewport( 0, 0, winWidth, winHeight );

	glClearColor(0.0,0.2,0.0,1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

// projection matrix for next rendering
	  glMatrixMode( GL_PROJECTION );

	  h = (GLfloat) winHeight / (GLfloat) winWidth;

	  znear = 0.3f;
	  zfar  = 5.0f;
	  xmax  = znear*1.0f;

	  glLoadIdentity();
	  glFrustum( -xmax, xmax, -xmax*h, xmax*h, znear, zfar );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

// construct inverseModel * view ...

// retrive viewPos in ModelSpace
	GLfloat invModelView[4][4];
	glGetFloatv(GL_MODELVIEW_MATRIX,(GLfloat*)invModelView);

// copy "translation part" ...

	glLoadIdentity();
// construct invView * model matrix

	// *= inverse view
    glTranslatef(0,0,-3.0);

	// *= model matrix
    glRotatef(angleY,0,1,0);
	glRotatef(angleX,1,0,0);

	glEnable(GL_DEPTH_TEST);

	glEnable(GL_CULL_FACE); 

	renderModel(modelCube);
//	renderSquareTex3D(gVolumeMapID,0.5);
}


/* new window size callback */
void GLFWCALL onReshape( int width, int height )
{
  GLfloat h = (GLfloat) height / (GLfloat) width;
  GLfloat xmax, znear, zfar;

  znear = 0.3f;
  zfar  = 4.0f;
  xmax  = znear*1.0f;

  glViewport( 0, 0, (GLint) width, (GLint) height );
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glFrustum( -xmax, xmax, -xmax*h, xmax*h, znear, zfar );

  winWidth=width;
  winHeight=height;
}

/* handle keyboard input */
void GLFWCALL onKey( int k, int action )
{
  if( action != GLFW_PRESS ) return;

  switch (k) {
    case GLFW_KEY_ESC:
        running = 0;
    break;
	case '+' : // increase max number of raysteps
		break;
	case '-' : // decrease max number of raysteps
		break;

    default:
        return;
  }
}

/* handle mouse move event */
void GLFWCALL onMouseMove( int x, int y )
{
    int button = GLFW_MOUSE_BUTTON_LEFT;
 /*
 GLFW_MOUSE_BUTTON_LEFT
 GLFW_MOUSE_BUTTON_RIGHT
 GLFW_MOUSE_BUTTON_MIDDLE
 */
    int state = glfwGetMouseButton( button );
    if (state == GLFW_PRESS)
    {
        fprintf(stderr,"%d %d\r",x,y);
    // process
    	angleY=x;
	    angleX=y;
    }
}

/* program entry */
int main(int argc, char *argv[])
{
    if( !glfwInit() )
    {
        fprintf( stderr, "Failed to initialize GLFW\n" );
        exit( EXIT_FAILURE );
    }
/*
glfwOpenWindow(int width, int height,
    int redbits, int greenbits, int bluebits, int alphabits,
    int depthbits, int stencilbits,
    int mode)
*/
    if( !glfwOpenWindow( winWidth,winHeight, 0,0,0,0, 16,0, GLFW_WINDOW ) )
    {
        fprintf( stderr, "Failed to open GLFW window\n" );
        glfwTerminate();
        exit( EXIT_FAILURE );
    }

    GLenum err = glewInit();
	if (err != GLEW_OK)
    {
		fprintf(stderr,"Error glew initialization failed (%s).\n",glewGetErrorString(err));
    }
    fprintf(stderr, "Status: Using GLEW %s\n", glewGetString(GLEW_VERSION));

    if (GLEW_VERSION_1_5)
    {
        fprintf(stderr, "[x] OpenGL version 1.5\n");
    }
    if (GLEW_VERSION_2_0)
    {
        fprintf(stderr, "[x] OpenGL version 2.0\n");
    }
    if (GLEW_VERSION_2_1)
    {
        fprintf(stderr, "[x] OpenGL version 2.1\n");
    }
    if (GLEW_VERSION_3_0)
    {
        fprintf(stderr, "[x] OpenGL version 3.0\n");
    }



    glfwSetWindowTitle( "cviko_10" );
    glfwEnable( GLFW_KEY_REPEAT );
    glfwSwapInterval( 1 ); // VSync 1x

    // Set callback functions
    glfwSetWindowSizeCallback( onReshape );

    glfwSetKeyCallback( onKey );
    glfwSetMousePosCallback( onMouseMove );

    onInit();

    // Main loop
    while( running )
    {
         onRender();
        // Swap buffers
        glfwSwapBuffers();
        glfwSleep( 0.01 );    // don't eat whole CPU time
        // Was the window closed?
        if( !glfwGetWindowParam( GLFW_OPENED ) )
        {
            running = 0;
        }
    }

    // Terminate GLFW
    glfwTerminate();

    // Exit program
    exit( EXIT_SUCCESS );
}

